'use strict';

const User = require.main.require('./src/user');
const Groups = require.main.require('./src/groups');
const db = require.main.require('./src/database');
const authenticationController = require.main.require('./src/controllers/authentication');

const crypto = require('crypto');
const util = require('util');
const nconf = module.parent.require('nconf');
const passport = module.parent.require('passport');
const Strategy = require('passport-strategy');

const MyInedo = {};

var preSharedKey;
var preSharedSecret;

function InedoPassportStrategy() {
	Strategy.call(this);
}
util.inherits(InedoPassportStrategy, Strategy);

InedoPassportStrategy.prototype.authenticate = function(req, options) {
	const self = this;

	if (!req.query.state || !req.query.payload) {
		return self.redirect('https://my.inedo.com/log-in/' + nconf.get('url_parsed').host + '?state=' + encodeURIComponent(options.state));
	}

	const state = req.query.state;
	const payload = Buffer.from(req.query.payload, 'hex');

	// payload contents:
	// (32 bytes) hmac-sha256 of data with pre-shared secret
	// (16 bytes) initialization vector
	// (multiple of 16 bytes, at least 3x16) aes-256-cbc-encrypted data with pre-shared key

	// fastest check: make sure the size makes sense
	if (payload.length < 32 + 16 + 3*16 || payload.length % 16 !== 0) {
		return self.fail();
	}

	const decipher = crypto.createDecipheriv('aes-256-cbc', preSharedKey, payload.slice(32, 32 + 16));
	const data = Buffer.concat([decipher.update(payload.slice(32 + 16)), decipher.final()]);
	const hmac = crypto.createHmac('sha256', preSharedSecret).update(data).digest();
	const signatureCorrect = crypto.timingSafeEqual(payload.slice(0, 32), hmac);
	const stateHash = crypto.createHash('sha256').update(state, 'utf8').digest();
	const stateCorrect = crypto.timingSafeEqual(data.slice(0, 32), stateHash);
	if (!signatureCorrect || !stateCorrect) {
		return self.fail();
	}

	// everything above this point must be constant-time for anything other than the opaque input size

	// data contents:
	// (32 bytes) sha256 of state (already checked at this point)
	// (8 bytes) unix timestamp (int64)
	// (1 byte) version number (0)
	// (remaining) MyInedo email address, utf-8 encoded

	const timestamp = data.readIntLE(32, 6) * 1000;
	if (data.readIntLE(32 + 6, 2) != 0) {
		// timestamp is too far in the future or past
		return self.error();
	}
	const timediff = Math.abs(Date.now() - timestamp);
	const version = data.readUInt8(32 + 8);

	// only allow payloads created within 1 minute of now (in either direction, to allow clock skew)
	if (timediff > 60000 || version !== 0) {
		return self.fail();
	}

	const email = data.toString('utf8', 32 + 8 + 1);
	return User.getUidByEmail(email, function (err, uid) {
		if (err) {
			return self.error(err);
		}

		if (uid !== null) {
			if (req.user && req.user.uid > 0 && req.user.uid !== uid) {
				return self.error(new Error('[[error:sso-multiple-association]]'));
			}

			// existing user
			return authenticationController.onSuccessfulLogin(req, uid, function(err) {
				if (err) {
					return self.error(err);
				}
				return self.success({uid: uid});
			});
		}

		var generatedUsername = email.substr(0, email.indexOf('@'));
		generatedUsername.replace(/(['"\s\-+.*[\]0-9\u00BF-\u1FFF\u2C00-\uD7FF\w]+)/g, '_');
		generatedUsername += '_' + '####'.replace(/#/g, function() {
			return Math.floor(Math.random() * 10).toString();
		});

		return User.create({
			username: generatedUsername,
			email: email
		}, function(err, uid) {
			if (err) {
				return self.error(err);
			}

			return self.success({uid: uid});
		});
	});
}

passport.use('inedo', new InedoPassportStrategy());

MyInedo.init = function(strategies, callback) {
	preSharedKey = Buffer.from(nconf.get('myinedo_ekey'), 'hex');
	preSharedSecret = Buffer.from(nconf.get('myinedo_skey'), 'hex');

	strategies.push({
		name: 'inedo',
		url: '/auth/inedo',
		callbackURL: '/auth/inedo/callback'
	});

	callback(null, strategies);
};

module.exports = MyInedo;
